using System;
using System.Collections.Generic;
using System.Text;

namespace CrossingOver
{
    class DoubleStrand
    {
        //at least how many letters must match in an area to be considered a match
        public const int MinMatchLength = 3;

        //strand A is assumed to be as long or longer than strand B.
        public string A, B;
        public static Random rand = new Random();

        public int[] Matching;

        //Ctor
        public DoubleStrand(string a, string b)
        {
            if (b.Length > a.Length)
            {
                B = a;
                A = b;
            }
            else
            {
                B = b;
                A = a;
            }

            Matching = new int[A.Length];
            for(int x = 0; x < A.Length; ++x)
            {
                Matching[x] = -1;
            }
        }

        public void Match()
        {
            //Seed the recursion
            MatchLongestSequence(0, A.Length, 0, B.Length);
        }

        /// <summary>
        /// Finds the longest string shared between a and b
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        string FindLongestMatch(string a, string b)
        {
            string ReturnMe = "";

            for(int x = 0; x < A.Length; ++x)
                for (int length = ReturnMe.Length + 1; x + length <= a.Length; ++length)
                    if (length > ReturnMe.Length && b.Contains(a.Substring(x, length)))
                        ReturnMe = a.Substring(x, length);

            return ReturnMe;
        }

        /// <summary>
        /// Returns the indexes of all times in string a that Match is found
        /// </summary>
        /// <param name="a"></param>
        /// <param name="Match"></param>
        /// <returns></returns>
        List<int> FindMatches(string a, string Match)
        {
            int offsetA = 0;
            List<int> ReturnMe = new List<int>();

            while (a.Contains(Match))
            {
                ReturnMe.Add(a.IndexOf(Match) + offsetA);
                a = a.Substring(a.IndexOf(Match) + Match.Length);
                offsetA = ReturnMe[ReturnMe.Count - 1] + 1;
            }

            return ReturnMe;
        }

        /// <summary>
        /// Matches the longest sequence in the sub strings, then calls this function again
        /// recursively on the left overs.
        /// </summary>
        /// <param name="StartOffsetInA"></param>
        /// <param name="LengthInA"></param>
        /// <param name="StartOffsetInB"></param>
        /// <param name="LengthInB"></param>
        void MatchLongestSequence(int StartOffsetInA, int LengthInA, int StartOffsetInB, int LengthInB)
        {
            string ASegment = A.Substring(StartOffsetInA, LengthInA);
            string BSegment = B.Substring(StartOffsetInB, LengthInB);

            string LongestMatch = FindLongestMatch(ASegment, BSegment);

            //There aren't enough matchings left to be considered a true match
            if (LongestMatch.Length < MinMatchLength)
                return;

            List<int> aList, bList;

            aList = FindMatches(ASegment, LongestMatch);
            bList = FindMatches(BSegment, LongestMatch);

            int BestDistance = int.MaxValue;
            int AStart = -1;
            int BStart = -1;

            foreach(int a in aList)
                foreach (int b in bList)
                {
                    int DistanceUp = Math.Abs(a - b);
                    int ARemaining = ASegment.Length - a - LongestMatch.Length + 1;
                    int BRemaining = BSegment.Length - b - LongestMatch.Length + 1;
                    int DistanceDown = Math.Abs(ARemaining - BRemaining);

                    int TotalDistance = DistanceUp + DistanceDown;
                    if (TotalDistance < BestDistance)
                    {
                        AStart = a;
                        BStart = b;
                        BestDistance = TotalDistance;
                    }
                }

            for (int x = 0; x < LongestMatch.Length; ++x)
                Matching[AStart + StartOffsetInA + x] = BStart + StartOffsetInB + x;

            MatchLongestSequence(StartOffsetInA, AStart, StartOffsetInB, BStart);
            MatchLongestSequence(StartOffsetInA + AStart + LongestMatch.Length, LengthInA - AStart - LongestMatch.Length,
                                 StartOffsetInB + BStart + LongestMatch.Length, LengthInB - BStart - LongestMatch.Length);

        }

        /// <summary>
        /// Performs a single cross over event.
        /// </summary>
        public void CrossOver()
        {
            bool AtLeastOnePlace = false;
            foreach (int x in Matching)
                if (x > 0)
                    AtLeastOnePlace = true;

            //Early out, no places to cross over.
            if (!AtLeastOnePlace)
                return;

            int RandPlace = rand.Next(Matching.Length);
            while (Matching[RandPlace] < 0)
                RandPlace = rand.Next(Matching.Length);

            //Swap downstreams of A and B:
            string Adown = A.Substring(RandPlace);
            string Bdown = B.Substring(Matching[RandPlace]);

            A = A.Remove(RandPlace);
            B = B.Remove(Matching[RandPlace]);

            A += Bdown;
            B += Adown;
        }

        public override string ToString()
        {
            String ReturnMe = "";

            //Draw the string of strand A that isn't matched.
            for(int x = 0; x < Matching.Length; ++x)
            {
                if(Matching[x] < 0)
                    ReturnMe += A[x];
                else
                    ReturnMe += " ";
            }

            ReturnMe += "\n";

            //Draw the string of strand A that is matched.
            for (int x = 0; x < Matching.Length; ++x)
            {
                if (Matching[x] >= 0)
                    ReturnMe += A[x];
                else
                    ReturnMe += "-";
            }

            ReturnMe += "\n";
            
            //Draw the string of strand B that is matched.
            for (int x = 0; x < Matching.Length; ++x)
            {
                if (Matching[x] >= 0)
                    ReturnMe += B[Matching[x]];
                else
                    ReturnMe += "-";
            }

            ReturnMe += "\n";

            return ReturnMe;
        }
    }
    class Program
    {
        static void Main(string[] args)
        {
            string A, B;            

            Console.WriteLine("Enter in the first string to use:");
            A = Console.ReadLine();
            Console.WriteLine("Enter in the second string to use:");
            B = Console.ReadLine();

            DoubleStrand strand = new DoubleStrand(A, B);

            strand.Match();

            Console.WriteLine(strand.ToString());

            Console.WriteLine("Enter in the number of crossover events.");
            int CrossoverEvents = int.Parse(Console.ReadLine());

            for (int x = 0; x < CrossoverEvents; ++x)
                strand.CrossOver();

            Console.WriteLine("The new strands:");
            Console.WriteLine(strand.A);
            Console.WriteLine(strand.B);

            Console.ReadLine();
        }
    }
}
